@@ -947,7 +947,10 @@ class TypeEngine(typing.Generic[T]):
             for arg in annotate_args:
                 if isinstance(arg, Renderable):
                     return arg.to_html(python_val)
-        return transformer.to_html(ctx, python_val, expected_python_type)
+        try:
+            return transformer.to_html(ctx, python_val, expected_python_type)
+        except:
+            return "unable to convert value to html string"
 
     @classmethod
     def named_tuple_to_variable_map(cls, t: typing.NamedTuple) -> _interface_models.VariableMap:
