@@ -1,10 +1,12 @@
+import base64
+import hashlib
 import math
 import mimetypes
 import os as _os
 import time
 import traceback
 from queue import Empty, Queue
-from threading import Event, Thread
+from threading import Thread
 from typing import List, Optional
 from urllib.parse import urlparse
 
@@ -26,6 +28,7 @@ class ThreadPool:
     def __init__(self, num_threads):
         self.tasks: Queue = Queue(num_threads)
         self.threads: List[Worker] = []
+        self.cache_accum = bytearray([0] * 32)
         for _ in range(num_threads):
             self.threads.append(Worker(self.tasks))
 
@@ -37,6 +40,9 @@ class ThreadPool:
         for t in self.threads:
             t.join()
 
+            for i in range(len(self.cache_accum)):
+                self.cache_accum[i] ^= t.cache_accum[i]
+
     def add_task(self, func, *args, **kargs):
         """Add a task to the queue"""
         self.tasks.put((func, args, kargs))
@@ -51,8 +57,7 @@ class ThreadPool:
         try:
             self.tasks.join()
         finally:
-            for t in self.threads:
-                t.join()
+            self.join()
 
 
 class Worker(Thread):
@@ -62,6 +67,7 @@ class Worker(Thread):
         Thread.__init__(self)
         self.tasks = tasks
         self.daemon = True
+        self.cache_accum = bytearray([0] * 32)
 
     def run(self):
         while True:
@@ -74,7 +80,13 @@ class Worker(Thread):
                 attempts = 0
                 while True:
                     try:
-                        func(*args, **kargs)
+                        res = func(*args, **kargs)
+
+                        if res["cache"] is not None:
+                            hash = hashlib.sha256(res["cache"].encode("utf-8")).digest()
+                            for i in range(len(self.cache_accum)):
+                                self.cache_accum[i] ^= hash[i]
+
                         break
                     except Exception as e:
                         attempts += 1
@@ -161,30 +173,40 @@ class LatchPersistence(DataPersistence):
         :param str remote_path: remote latch:// path
         :param str local_path: directory to copy to
         """
-        r = _session.post(
-            self._latch_endpoint + "/api/get-presigned-urls-for-dir",
-            json={"object_url": remote_path, "execution_name": _os.environ.get("FLYTE_INTERNAL_EXECUTION_ID")},
-            timeout=90,
-        )
-        if r.status_code != 200:
-            raise FlyteUserException("failed to download `{}`".format(remote_path))
+        internal_execution_id = _os.environ.get("FLYTE_INTERNAL_EXECUTION_ID")
 
-        dir_key = self._split_s3_path_to_key(remote_path)[1:]
-        dir_key = _enforce_trailing_slash(dir_key)
-        key_to_url_map = r.json()["key_to_url_map"]
+        if internal_execution_id is not None:
+            r = _session.post(
+                self._latch_endpoint + "/api/get-presigned-urls-for-dir",
+                json={"object_url": remote_path, "execution_name": internal_execution_id},
+                timeout=90,
+            )
+            if r.status_code != 200:
+                raise FlyteUserException("failed to download `{}`".format(remote_path))
 
-        task_tuples = []
-        for key, url in key_to_url_map.items():
-            local_file_path = _os.path.join(local_path, key.replace(dir_key, "", 1))
-            dir = "/".join(local_file_path.split("/")[:-1])
-            _os.makedirs(dir, exist_ok=True)
-            task_tuples.append((url, local_file_path))
+            dir_key = self._split_s3_path_to_key(remote_path)[1:]
+            dir_key = _enforce_trailing_slash(dir_key)
+            key_to_url_map = r.json()["key_to_url_map"]
 
-        pool = ThreadPool(100)
-        pool.map(get, task_tuples)
-        pool.start()
-        pool.wait_completion()
-        return True
+            task_tuples = []
+            for key, url in key_to_url_map.items():
+                local_file_path = _os.path.join(local_path, key.replace(dir_key, "", 1))
+                dir = "/".join(local_file_path.split("/")[:-1])
+                _os.makedirs(dir, exist_ok=True)
+                task_tuples.append((url, local_file_path))
+
+            pool = ThreadPool(100)
+            pool.map(get, task_tuples)
+            pool.start()
+            pool.wait_completion()
+            return True
+        else:
+            try:
+                from latch_cli.services.cp import cp
+
+                cp(remote_path, local_path)
+            except Exception as e:
+                raise FlyteUserException("failed to get presigned urls for `{}`".format(remote_path)) from e
 
     def download(self, remote_path: str, local_path: str) -> bool:
         """
@@ -214,11 +236,11 @@ class LatchPersistence(DataPersistence):
                 raise FlyteUserException("failed to get presigned url for `{}`".format(remote_path)) from e
 
     @staticmethod
-    def __upload(args) -> bool:
+    def __upload(args):
         return LatchPersistence._upload(args[0], args[1], args[2], args[3])
 
     @staticmethod
-    def _upload(file_path: str, to_path: str, chunk_size: int, endpoint: str) -> bool:
+    def _upload(file_path: str, to_path: str, chunk_size: int, endpoint: str):
         file_size = _os.path.getsize(file_path)
         nrof_parts = math.ceil(float(file_size) / chunk_size)
         content_type = mimetypes.guess_type(file_path)[0]
@@ -269,7 +291,9 @@ class LatchPersistence(DataPersistence):
             )
             if r.status_code != 200:
                 raise RuntimeError("failed to complete upload for `{}`".format(to_path))
-            return True
+
+            data = r.json()
+            return {"cache": data["VersionId"]}
         else:
             try:
                 from latch_cli.services.cp import cp
@@ -278,14 +302,14 @@ class LatchPersistence(DataPersistence):
             except Exception as e:
                 raise FlyteUserException("failed to get presigned upload urls for `{}`".format(to_path)) from e
 
-    def upload(self, file_path: str, to_path: str) -> bool:
+    def upload(self, file_path: str, to_path: str):
         """
         :param str file_path:
         :param str to_path:
         """
         return LatchPersistence._upload(file_path, to_path, self._chunk_size, self._latch_endpoint)
 
-    def upload_directory(self, local_path: str, remote_path: str) -> bool:
+    def upload_directory(self, local_path: str, remote_path: str):
         """
         :param str local_path:
         :param str remote_path:
@@ -310,7 +334,9 @@ class LatchPersistence(DataPersistence):
         pool.map(LatchPersistence.__upload, task_tuples)
         pool.start()
         pool.wait_completion()
-        return True
+
+        cache = base64.standard_b64encode(bytes(pool.cache_accum)).decode("ascii")
+        return {"cache": cache}
 
     def construct_path(self, add_protocol: bool, add_prefix: bool, *paths: str) -> str:
         paths = list(paths)  # make type check happy
